from django.contrib import admin
from django.contrib.auth import views as auth_views
from axes.decorators import watch_login

class FailedLoginMiddleware(object):

    def __init__(self, *args, **kwargs):
        super(FailedLoginMiddleware, self).__init__(*args, **kwargs)

        # watch the admin login page
        admin.site.login = watch_login(admin.site.login)

        # and the regular auth login page
        auth_views.login = watch_login(auth_views.login)



from axes.models import AccessAttempt

def reset(ip=None, silent=False):
    if not ip:
        attempts = AccessAttempt.objects.all()
        if attempts:
            attempts.delete()
        else:
            if not silent:
                print 'No attempts found.'
    else:
        try:
            attempt = AccessAttempt.objects.get(ip_address=ip)
        except:
            if not silent:
                print 'No matching attempt found.'
        else:
            attempt.delete()




from django.db import models
from django.conf import settings

FAILURES_DESC = 'Failed Logins'

#XXX TODO
# set unique by user_agent, ip
# make user agent, ip indexed fields

class AccessAttempt(models.Model):
    user_agent = models.CharField(max_length=255)
    ip_address = models.IPAddressField('IP Address')
    get_data = models.TextField('GET Data')
    post_data = models.TextField('POST Data')
    http_accept = models.CharField('HTTP Accept', max_length=255)
    path_info = models.CharField('Path', max_length=255)
    failures_since_start = models.PositiveIntegerField(FAILURES_DESC)
    attempt_time = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        return u'Attempted Access: %s' % self.attempt_time

    def failures(self):
        return self.failures_since_start

    class Meta:
        ordering = ['-attempt_time']



VERSION = (1, 0, 0, 'rc2')

def get_version():
    return '%s.%s.%s-%s' % VERSION

try:
    from django.conf import settings
    import logging, os

    LOGFILE = os.path.join(settings.DIRNAME, 'axes.log')
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt='%a, %d %b %Y %H:%M:%S',
                        filename=LOGFILE,
                        filemode='w')

    fileLog = logging.FileHandler(LOGFILE, 'w')
    fileLog.setLevel(logging.DEBUG)

    # set a format which is simpler for console use
    formatter = logging.Formatter('%(asctime)s %(name)-12s: %(levelname)-8s %(message)s')

    # tell the handler to use this format
    fileLog.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger('').addHandler(fileLog)
except:
    # if we have any problems, we most likely don't have a settings module loaded
    pass



from django.contrib import admin
from axes.models import AccessAttempt

class AccessAttemptAdmin(admin.ModelAdmin):
    list_display = ('attempt_time', 'ip_address', 'user_agent', 'path_info', 'failures_since_start')
    list_filter = ['attempt_time', 'ip_address', 'path_info']
    search_fields = ['ip_address', 'user_agent', 'path_info']
    date_hierarchy = 'attempt_time'
    fieldsets = (
        (None, {
            'fields': ('path_info', 'failures_since_start')
        }),
        ('Form Data', {
            'fields': ('get_data', 'post_data')
        }),
        ('Meta Data', {
            'fields': ('user_agent', 'ip_address', 'http_accept')
        })
    )

admin.site.register(AccessAttempt, AccessAttemptAdmin)


from datetime import datetime, timedelta
import logging

from django.conf import settings
from django.contrib.auth import logout
from django.core.cache import cache
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template import RequestContext

from axes.models import AccessAttempt
import axes

# see if the user has overridden the failure limit
FAILURE_LIMIT = getattr(settings, 'AXES_LOGIN_FAILURE_LIMIT', 3)

# see if the user has set axes to lock out logins after failure limit
LOCK_OUT_AT_FAILURE = getattr(settings, 'AXES_LOCK_OUT_AT_FAILURE', True)

USE_USER_AGENT = getattr(settings, 'AXES_USE_USER_AGENT', False)

COOLOFF_TIME = getattr(settings, 'AXES_COOLOFF_TIME', None)
if isinstance(COOLOFF_TIME, int):
    COOLOFF_TIME = timedelta(hours=COOLOFF_TIME)

LOGGER = getattr(settings, 'AXES_LOGGER', 'axes.watch_login')

LOCKOUT_TEMPLATE = getattr(settings, 'AXES_LOCKOUT_TEMPLATE', None)
LOCKOUT_URL = getattr(settings, 'AXES_LOCKOUT_URL', None)
VERBOSE = getattr(settings, 'AXES_VERBOSE', True)

def query2str(items):
    return '\n'.join(['%s=%s' % (k, v) for k,v in items])

log = logging.getLogger(LOGGER)
if VERBOSE:
    log.info('AXES: BEGIN LOG')
    log.info('Using django-axes ' + axes.get_version())

def get_user_attempt(request):
    """
    Returns access attempt record if it exists.
    Otherwise return None.
    """
    ip = request.META.get('REMOTE_ADDR', '')
    if USE_USER_AGENT:
        ua = request.META.get('HTTP_USER_AGENT', '<unknown>')
        attempts = AccessAttempt.objects.filter(
            user_agent=ua,
            ip_address=ip
        )
    else:
        attempts = AccessAttempt.objects.filter(
            ip_address=ip
        )

    if not attempts:
        return None

    attempt = attempts[0]
    if COOLOFF_TIME and attempt.attempt_time + COOLOFF_TIME < datetime.now():
        attempt.delete()
        return None

    return attempt

def watch_login(func):
    """
    Used to decorate the django.contrib.admin.site.login method.
    """

    def decorated_login(request, *args, **kwargs):
        # share some useful information
        if func.__name__ != 'decorated_login' and VERBOSE:
            log.info('AXES: Calling decorated function: %s' % func.__name__)
            if args: log.info('args: %s' % args)
            if kwargs: log.info('kwargs: %s' % kwargs)

        # call the login function
        response = func(request, *args, **kwargs)

        if func.__name__ == 'decorated_login':
            # if we're dealing with this function itself, don't bother checking
            # for invalid login attempts.  I suppose there's a bunch of
            # recursion going on here that used to cause one failed login
            # attempt to generate 10+ failed access attempt records (with 3
            # failed attempts each supposedly)
            return response

        if request.method == 'POST':
            # see if the login was successful
            login_unsuccessful = (
                response and
                not response.has_header('location') and
                response.status_code != 302
            )
            if check_request(request, login_unsuccessful):
                return response

            if LOCKOUT_TEMPLATE:
                context = RequestContext(request, {
                    'cooloff_time': COOLOFF_TIME,
                    'failure_limit': FAILURE_LIMIT,
                })
                return render_to_response(LOCKOUT_TEMPLATE, context)

            if LOCKOUT_URL:
                return HttpResponseRedirect(LOCKOUT_URL)

            if COOLOFF_TIME:
                return HttpResponse("Account locked: too many login attempts.  "
                                    "Please try again later.")
            else:
                return HttpResponse("Account locked: too many login attempts.  "
                                    "Contact an admin to unlock your account.")
        return response

    return decorated_login

def check_request(request, login_unsuccessful):
    failures = 0
    attempt = get_user_attempt(request)

    if attempt:
        failures = attempt.failures_since_start

    # no matter what, we want to lock them out
    # if they're past the number of attempts allowed
    if failures > FAILURE_LIMIT and LOCK_OUT_AT_FAILURE:
        # We log them out in case they actually managed to enter
        # the correct password.
        logout(request)
        log.warn('AXES: locked out %s after repeated login attempts.' %
                 attempt.ip_address)
        return False

    if login_unsuccessful:
        # add a failed attempt for this user
        failures += 1

        # Create an AccessAttempt record if the login wasn't successful
        # has already attempted, update the info
        if attempt:
            attempt.get_data = '%s\n---------\n%s' % (
                attempt.get_data,
                query2str(request.GET.items()),
            )
            attempt.post_data = '%s\n---------\n%s' % (
                attempt.post_data,
                query2str(request.POST.items())
            )
            attempt.http_accept = request.META.get('HTTP_ACCEPT', '<unknown>')
            attempt.path_info = request.META.get('PATH_INFO', '<unknown>')
            attempt.failures_since_start = failures
            attempt.attempt_time = datetime.now()
            attempt.save()
            log.info('AXES: Repeated login failure by %s. Updating access '
                     'record. Count = %s' %
                     (attempt.ip_address, failures))
        else:
            ip = request.META.get('REMOTE_ADDR', '')
            ua = request.META.get('HTTP_USER_AGENT', '<unknown>')
            attempt = AccessAttempt.objects.create(
                user_agent=ua,
                ip_address=ip,
                get_data=query2str(request.GET.items()),
                post_data=query2str(request.POST.items()),
                http_accept=request.META.get('HTTP_ACCEPT', '<unknown>'),
                path_info=request.META.get('PATH_INFO', '<unknown>'),
                failures_since_start=failures
            )
            log.info('AXES: New login failure by %s. Creating access record.' %
                     ip)
    else:
        # user logged in -- forget the failed attempts
        if attempt:
            attempt.delete()

    return True



from django.test import TestCase, Client
from django.conf import settings
from django.contrib import admin
import random
from django.contrib.auth.models import User

from models import AccessAttempt
from decorators import FAILURE_LIMIT

# Only run tests if they have axes in middleware

# Basically a functional test
class AccessAttemptTest(TestCase):
    NOT_GONNA_BE_PASSWORD = "sfdlermmvnLsefrlg0c9gjjPxmvLlkdf2#"
    NOT_GONNA_BE_USERNAME = "whywouldyouohwhy"

    def setUp(self):
        for i in range(0, random.randrange(10, 50)):
            username = "person%s" % i
            email = "%s@example.org" % username
            u = User.objects.create_user(email=email, username=username)
            u.is_staff = True
            u.save()

    def _gen_bad_password(self):
        return AccessAttemptTest.NOT_GONNA_BE_PASSWORD + str(random.random())

    def _random_username(self, correct_username=False):
        if not correct_username:
            return (AccessAttemptTest.NOT_GONNA_BE_USERNAME +
                    str(random.random()))[:30]
        else:
            return random.choice(User.objects.filter(is_staff=True))

    def _attempt_login(self, correct_username=False, user=""):
        response = self.client.post(
        '/admin/', {'username': self._random_username(correct_username),
                    'password': self._gen_bad_password()}
         )
        return response

    def test_login_max(self, correct_username=False):
        for i in range(0, FAILURE_LIMIT):
            response = self._attempt_login(correct_username=correct_username)
            self.assertContains(response, "this_is_the_login_form")
        # So, we shouldn't have gotten a lock-out yet.
        # But we should get one now
        response = self._attempt_login()
        self.assertContains(response, "Account locked")

    def test_login_max_with_more(self, correct_username=False):
        for i in range(0, FAILURE_LIMIT):
            response = self._attempt_login(correct_username=correct_username)
            self.assertContains(response, "this_is_the_login_form")
        # So, we shouldn't have gotten a lock-out yet.
        # But we should get one now
        for i in range(0, random.randrange(1, 100)):
            # try to log in a bunch of times
            response = self._attempt_login()
            self.assertContains(response, "Account locked")

    def test_with_real_username_max(self):
        self.test_login_max(correct_username=True)

    def test_with_real_username_max_with_more(self):
        self.test_login_max_with_more(correct_username=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import axes
import sys, os

setup(
    name='django-axes',
    version=axes.get_version(),
    description="Keep track of failed login attempts in Django-powered sites.",
    long_description=open('README.rst', 'r').read(),
    keywords='django, security, authentication',
    author='Josh VanderLinden, Philip Neustrom',
    author_email='codekoala@gmail.com',
    url='http://bitbucket.org/codekoala/django-axes/',
    license='MIT',
    package_dir={'axes': 'axes'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: Log Analysis',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Security',
        'Topic :: System :: Logging',
    ]
)
